/*
Copyright (c) 2017, Intel Corp.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#include <linux/module.h>

#define __NO_VERSION__
#include <linux/version.h>

#if LINUX_VERSION_CODE > KERNEL_VERSION(4,11,99)
#include <linux/uaccess.h>		// for copy_from_user
#else
#include <asm/uaccess.h>		// for copy_from_user
#endif

#include <asm/io.h>				// for inb and outb
//#include <asm/system.h>			// for global save_flag

#include <linux/spinlock.h>		// for spinlock
#include <linux/ioport.h>		// for check_region
#include <linux/rtc.h>			// for /dev/rtc interface
#include <linux/mc146818rtc.h>	// for rtc spin lock

/********************************************************************************
 * header for driver															*
 ********************************************************************************/
#include "wrapper.h"
#include "ioctl.h"

/********************************************************************************
 * Local Function Prototypes													*
 ********************************************************************************/
// CMOS API Interface
static BOOL CmosKernelRead(unsigned long cmosapiInfo);
static BOOL CmosKernelWrite(unsigned long cmosapiInfo);
static BOOL CmosTestRead(unsigned long cmosapiInfo);
static BOOL CmosTestWrite(unsigned long cmosapiInfo);

/********************************************************************************
 * Global Function Prototypes													*
 ********************************************************************************/
extern spinlock_t g_iommapSpinLock;

/********************************************************************************
 * Global variable																*
 ********************************************************************************/
const static UINT8 PORT0 = 0x70;
const static UINT8 DATA0 = 0x71;
const static UINT8 PORT1 = 0x72;
const static UINT8 DATA1 = 0x73;

/********************************************************************************
 * Function	: CmosTestDispatch													*
 * Access	: Global															*
 * Abstract	: Dispatcher function for all CMOS access.							*
 ********************************************************************************/
BOOL CmosTestDispatch(unsigned int cmosapiIoctl, unsigned long cmosapiInfo)
{
	int status = EINVAL;

	// Switch according to the ioctl called
	switch (cmosapiIoctl) {

	// For CMOS API, reserve 0x60 to 0x6F
	case CMOSAPI_KERNEL_READ:	// Do not use this for new code!
		status = CmosKernelRead(cmosapiInfo);	
		break;

	case CMOSAPI_KERNEL_WRITE:	// Do not use this for new code!
		status = CmosKernelWrite(cmosapiInfo);	
		break;

	case CMOSAPI_READ:	
		status = CmosTestRead(cmosapiInfo);	
		break;

	case CMOSAPI_WRITE:	
		status = CmosTestWrite(cmosapiInfo);	
		break;

	default:
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID CASE FOR CMOS ACCESS, CMD(%d)\n",__FUNCTION__, __LINE__, cmosapiIoctl);
        return EINVAL;
  }

  return status;

}

/********************************************************************************
 * Function	: CmosTestRead														*
 * Access	: Local																*
 * Abstract	: CMOS read ioctl call.						 						*
 ********************************************************************************/
static BOOL CmosTestRead(unsigned long cmosapiInfo)
{
	int i, status; unsigned long flags;

	CMOS_REQUEST Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, cmosapiInfo);

	status = copy_from_user( &Block, (void*)cmosapiInfo, sizeof(CMOS_REQUEST));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// Check that an the read does not exceed address 0xFF.
	if(Block.uStart + Block.uLength > 0x100) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID LENGTH(%d)\n",__FUNCTION__, __LINE__, Block.uStart + Block.uLength);
		return EOVERFLOW;
	}

	// zero the output buffer
	memset(Block.uData,0,0xFF);

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	// if standard 128 byte CMOS
	if((Block.uStart < 128) && ((Block.uStart + Block.uLength) < 128)) {
		for(i = 0; (unsigned long)i < Block.uLength; i++) {
			outb(Block.uStart + i, PORT0);
			Block.uData[i] = (UINT8)inb(DATA0);
		}
	} 

	// if extended 256 byte CMOS
	if((Block.uStart < 128) && ((Block.uStart + Block.uLength) >= 128)) {
		int LowLength  = 128 - Block.uStart;
		int HighLength = Block.uLength - LowLength;

		// read port 0x70 and 0x71 for low bank
		for(i = 0; i < LowLength; i++) {
			outb(Block.uStart + i, PORT0);
			Block.uData[i] = (UINT8)inb(DATA0);
		}
		
		// read port 0x72 and 0x73 for high bank
		for(i = 0; i < HighLength; i++) {
			outb(i, PORT1);
			Block.uData[LowLength + i] = (UINT8)inb(DATA1);
		}
	} 

	// if extended 256 byte CMOS
	if(Block.uStart >= 128) {
		for(i = 0; (unsigned long)i < Block.uLength; i++) {
			outb((Block.uStart & 0x7F) + i, PORT1);
			Block.uData[i] = (UINT8)inb(DATA1);
		}	
	}

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/ }while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)cmosapiInfo, &Block, (sizeof(CMOS_REQUEST)));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}


#ifdef _DEBUG
	DMSG (KERN_INFO "Block.uStart 0x%X, Block.uLength 0x%X\n", (unsigned)Block.uStart, (unsigned)Block.uLength);

	for(i = 0; (unsigned long)i < Block.uLength; i++)
		DMSG (KERN_INFO "Block.uData[0x%X] 0x%X\n",i,(unsigned)Block.uData[i]);
#endif

    
	return SUCCESSFUL;
}

/********************************************************************************
 * Function	: CmosTestWrite														*
 * Access	: Local																*
 * Abstract	: CMOS write ioctl call.						 					*
 ********************************************************************************/
static BOOL CmosTestWrite(unsigned long cmosapiInfo)
{
	int i, status; unsigned long flags;

	CMOS_REQUEST Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, cmosapiInfo);

	status = copy_from_user( &Block, (void*)cmosapiInfo, sizeof(CMOS_REQUEST));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// Check that an the read does not exceed address 0xFF.
	if(Block.uStart + Block.uLength > 0x100) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID LENGTH(%d)\n",__FUNCTION__, __LINE__, Block.uStart + Block.uLength);
		return EOVERFLOW;
	}

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	// if standard 128 byte CMOS
	if((Block.uStart < 128) && ((Block.uStart + Block.uLength) < 128)) {
		for(i = 0; (unsigned long)i < Block.uLength; i++) {
			outb(Block.uStart + i, PORT0);
			outb(Block.uData[i], DATA0);
		}
	} 

	// if extended 256 byte CMOS
	if((Block.uStart < 128) && ((Block.uStart + Block.uLength) >= 128)) {
		int LowLength  = 128 - Block.uStart;
		int HighLength = Block.uLength - LowLength;

		// read port 0x70 and 0x71 for low bank
		for(i = 0; i < LowLength; i++) {
			outb(Block.uStart + i, PORT0);
			outb(Block.uData[i], DATA0);
		}
		
		// read port 0x72 and 0x73 for high bank
		for(i = 0; i < HighLength; i++) {
			outb(i, PORT1);
			outb(Block.uData[LowLength + i], DATA1);
		}
	} 

	// if extended 256 byte CMOS
	if(Block.uStart >= 128) {
		for(i = 0; (unsigned long)i < Block.uLength; i++) {
			outb((Block.uStart & 0x7F) + i, PORT1);
			outb(Block.uData[i], DATA1);
		}	
	}

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/ }while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);


#ifdef _DEBUG
	DMSG (KERN_INFO "Block.uStart 0x%X, Block.uLength 0x%X\n", (unsigned)Block.uStart, (unsigned)Block.uLength);

	for(i = 0; (unsigned long)i < Block.uLength; i++)
		DMSG (KERN_INFO "Block.uData[0x%X] 0x%X\n",i,(unsigned)Block.uData[i]);
#endif

    return SUCCESSFUL;
}





////////////////////////////////////////////////////
// below function is provided by /dev/rtc interface.
// there is a global rtc_lock provided by kernel to
// update the RTC value as well.
////////////////////////////////////////////////////


/********************************************************************************
 * Function	: CmosKernelRead													*
 * Access	: Local																*
 * Abstract	: CMOS read ioctl call thru /dev/rtc interface.	 					*
 ********************************************************************************/
static BOOL CmosKernelRead(unsigned long cmosapiInfo)
{

#ifdef i386
	int i, status; unsigned long flags;

	CMOS_REQUEST Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, cmosapiInfo);

	status = copy_from_user( &Block, (void*)cmosapiInfo, sizeof(CMOS_REQUEST));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// Check that an the read does not exceed address 0xFF.
	if(Block.uStart + Block.uLength > 0x100) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID LENGTH(%d)\n",__FUNCTION__, __LINE__, Block.uStart + Block.uLength);
		return EOVERFLOW;
	}

	// zero the output buffer
	memset(Block.uData,0,0xFF);

	//below are provided by kernel in mc146818rtc.h
	spin_lock_irqsave(&rtc_lock, flags);

	// if standard 128 byte CMOS
	if((Block.uStart < 128) && ((Block.uStart + Block.uLength) < 128)) {
		for(i = 0; (unsigned long)i < Block.uLength; i++)
			Block.uData[i] = CMOS_READ(Block.uStart+i);
	} 
	
	spin_unlock_irqrestore(&rtc_lock, flags);

	status = copy_to_user((void*)cmosapiInfo, &Block, (sizeof(CMOS_REQUEST)));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}


#ifdef _DEBUG
	DMSG(KERN_INFO "Block.uStart 0x%X, Block.uLength 0x%X\n", (unsigned)Block.uStart, (unsigned)Block.uLength);

	for(i = 0; (unsigned long)i < Block.uLength; i++)
		DMSG(KERN_INFO "Block.uData[0x%X] 0x%X\n",i,(unsigned)Block.uData[i]);
#endif
    
	return SUCCESSFUL;
#else
	return CmosTestRead(cmosapiInfo);
#endif
}

/********************************************************************************
 * Function	: CmosKernelWrite													*
 * Access	: Local																*
 * Abstract	: CMOS write ioctl call thru /dev/rtc interface. 					*
 ********************************************************************************/
static BOOL CmosKernelWrite(unsigned long cmosapiInfo)
{
#ifdef i386
	int i, status; unsigned long flags;

	CMOS_REQUEST Block;

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, cmosapiInfo);

	status = copy_from_user( &Block, (void*)cmosapiInfo, sizeof(CMOS_REQUEST));
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// Check that an the read does not exceed address 0xFF.
	if(Block.uStart + Block.uLength > 0x100) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID LENGTH(%d)\n",__FUNCTION__, __LINE__, Block.uStart + Block.uLength);
		return EOVERFLOW;
	}

	//below are provided by kernel in mc146818rtc.h
	spin_lock_irqsave(&rtc_lock, flags);

	// if standard 128 byte CMOS
	if((Block.uStart < 128) && ((Block.uStart + Block.uLength) < 128)) {
		for(i = 0; (unsigned long)i < Block.uLength; i++)
			CMOS_WRITE(Block.uData[i], Block.uStart + i);
	} 

	spin_unlock_irqrestore(&rtc_lock, flags);


#ifdef _DEBUG
	DMSG(KERN_INFO "Block.uStart 0x%X, Block.uLength 0x%X\n", (unsigned)Block.uStart, (unsigned)Block.uLength);

	for(i = 0; (unsigned long)i < Block.uLength; i++)
		DMSG(KERN_INFO "Block.uData[0x%X] 0x%X\n",i,(unsigned)Block.uData[i]);
#endif

    return SUCCESSFUL;

#else
	return CmosTestWrite(cmosapiInfo);
#endif
}

