/*
Copyright (c) 2017, Intel Corp.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __KERNEL__
#define __KERNEL__
#endif

#ifndef MODULE
#define MODULE
#endif

#include <linux/kernel.h>
#include <linux/module.h>

#define __NO_VERSION__
#include <linux/version.h>

#if LINUX_VERSION_CODE > KERNEL_VERSION(4,11,99)
#include <linux/uaccess.h>		/* copy_to_user	*/
#else
#include <asm/uaccess.h>		/* copy_to_user	*/
#endif

//#include <asm/system.h>		// for global save_flag
#include <asm/io.h>			// for phys_to_virt
#include <asm/pgtable.h>	// for pte_offset_map

#include <linux/slab.h>		// for kmalloc
#include <linux/vmalloc.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/pagemap.h>
#include <linux/spinlock.h>	// for spin lock
#include <linux/sched.h>	// for init_mm
#include <linux/string.h>	// for memcpy

/********************************************************************************
 * header for driver															*
 ********************************************************************************/
#include "wrapper.h"
#include "ioctl.h"

/********************************************************************************
 * Local Function Prototypes													*
 ********************************************************************************/
// MEMORY API Interface
static BOOL MemMapTestGetPhysicalAddress(unsigned long memmapapiInfo);
//static BOOL MemMapTestMapVirtualAddress(unsigned long memmapapiInfo);

/********************************************************************************
 * Global Function Prototypes													*
 ********************************************************************************/
extern spinlock_t g_iommapSpinLock;

/********************************************************************************
 * Local variable																*
 ********************************************************************************/
static struct vm_area_struct *g_vma = NULL;
static unsigned long g_vma_offset = 0;

/********************************************************************************
 * Local Function Prototypes													*
 ********************************************************************************/
#ifdef _DEBUG
int driver_mmap(struct file *file, struct vm_area_struct *vma);
#endif

/********************************************************************************
 * Function	: MemMapTestDispatch												*
 * Access	: Global															*
 * Abstract	: Dispatcher function for all Memory access.						*
 ********************************************************************************/
BOOL MemMapTestDispatch(unsigned int memapiIoctl, unsigned long memmapapiInfo)
{
	int status = EINVAL;

	// Switch according to the ioctl called
	switch (memapiIoctl) {

	// For Memory API, reserve 0x70 to 0x7F
	case MEMMAPAPI_GET_PHYSICAL_ADDRESS:	
		status = MemMapTestGetPhysicalAddress(memmapapiInfo);	
		break;

	default:
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID CASE FOR MEMORY ACCESS, CMD(%d)\n",__FUNCTION__, __LINE__, memapiIoctl);
        return EINVAL;
  }

  return status;

}

/********************************************************************************
 * Function	: MemMapTestGetPhysicalAddress										*
 * Access	: Local																*
 * Abstract	: Memory physical address translation ioctl call.					*
 ********************************************************************************/
static BOOL MemMapTestGetPhysicalAddress(unsigned long memmapapiInfo)
{
	int status;	unsigned long flags;
	PHYSICAL_ADDRESS Block[2];
	struct vm_area_struct *vma = NULL;
	struct page *page = NULL;
	pgd_t *pgd = NULL;	
	pmd_t *pmd = NULL;	
	pte_t *ptep = NULL, pte;
	#if LINUX_VERSION_CODE > KERNEL_VERSION(4,11,99)
	p4d_t *p4d = NULL;
	#endif
	PUD_DEFINE

	DMSG("FUNCTION: %s, LINE: %d, ARG(0x%lX)\n", __FUNCTION__, __LINE__, memmapapiInfo);

	status = copy_from_user(&Block, (void*)memmapapiInfo, sizeof(PHYSICAL_ADDRESS)*2);
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_FROM_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	// Unsupported OS version
	CHECK_UNKNOW_OS_ERROR_1();

	// disable global irq and spinlock
	//do { save_flags(flags);	cli(); spin_lock(&g_iommapSpinLock); }while (0);
	//spin_lock_irqsave(&g_iommapSpinLock, flags);
	XSPIN_LOCK(g_iommapSpinLock, flags);

	// find the current vma value.
	//vma = find_vma(current->mm, Block[0].uLow);
	vma = find_vma(current->mm, (unsigned long)Block[0].uAddress);
	if(vma == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID VMA VALUE\n",__FUNCTION__, __LINE__);
		goto FailMem;
	}

	printk("<1>vma->vm_start 0x%.8lX vma->vm_end 0x%.8lX\n",vma->vm_start, vma->vm_end);
	printk("<1>vma->vm_pgoff 0x%.8lX vma->vm_flags 0x%.8lX\n",vma->vm_pgoff, vma->vm_flags);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	printk("<1>current->mm 0x%8p current->active_mm 0x%8p init_mm 0x%8p\n",
		current->mm, current->active_mm, &init_mm);
	printk("<1>current->mm->pgd 0x%8p current->active_mm->pgd 0x%8p init_mm.pgd 0x%8p\n",
		current->mm->pgd, current->active_mm->pgd, init_mm.pgd);
#else
	printk("<1>current->mm 0x%8p current->active_mm 0x%8p\n",
		current->mm, current->active_mm);
	printk("<1>current->mm->pgd 0x%8p current->active_mm->pgd 0x%8p\n",
		current->mm->pgd, current->active_mm->pgd);
#endif

	// init global vma value. g_vma and g_vma_offset will be used in mmap function for debug
	XINIT_VMA(g_vma, g_vma_offset);

	// physical RAM address translation is done by walking thru the page table.
	// walking thru 3 level of page directory.
	//pgd = pgd_offset(current->mm, Block[0].uLow);
	pgd = pgd_offset(current->mm, (unsigned long)Block[0].uAddress);
	//pgd = current->mm->pgd;
	if(pgd == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PGD VALUE\n",__FUNCTION__, __LINE__);
		goto FailMem;
	}

#if LINUX_VERSION_CODE > KERNEL_VERSION(4,11,99)
	p4d = p4d_offset(pgd, (unsigned long)Block[0].uAddress);
	PUD_OFFSET_WRAPPER(p4d, (unsigned long)Block[0].uAddress);
#else
	PUD_OFFSET_WRAPPER(pgd, (unsigned long)Block[0].uAddress);
#endif
	
	//pmd = pmd_offset(pgd, Block[0].uLow);
	//pmd = pmd_offset(pgd, (unsigned long)Block[0].uAddress);
	pmd = pmd_offset(PGD_OR_PUD, (unsigned long)Block[0].uAddress);
	if(pmd == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PMD VALUE\n",__FUNCTION__, __LINE__);
		goto FailMem;
	}

	//ptep = PTE_OFFSET_WRAPPER(pmd, Block[0].uLow);
	ptep = PTE_OFFSET_WRAPPER(pmd, (unsigned long)Block[0].uAddress);

	if(ptep == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PPTE VALUE\n",__FUNCTION__, __LINE__);
		goto FailMem;
	}

	pte = *ptep;

	PTE_UNMAP_WRAPPER(ptep);

	if(!pte_present(pte)) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PTE VALUE\n",__FUNCTION__, __LINE__);
		goto FailMem;
	}

	page = pte_page(pte);
	if(page == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PAGE VALUE\n",__FUNCTION__, __LINE__);
		goto FailMem;
	}

	printk("<1>pgd=0x%8p pmd=0x%8p pte=0x%8p page=0x%8p\n", pgd, pmd, ptep, page);

	// return the physical address. it will return 64bit is the kernel is compiled in HIGHMEM
	Block[1].uAddress = page_to_phys(page);

	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/ }while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);

	status = copy_to_user((void*)memmapapiInfo, &Block, sizeof(PHYSICAL_ADDRESS)*2);
	if(status != SUCCESSFUL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: COPY_TO_USER(%d)\n",__FUNCTION__, __LINE__, status);
		return EFAULT;
	}

	goto PassMem;

FailMem:
	// restore global irq and unlock
	//do { spin_unlock(&g_iommapSpinLock); restore_flags(flags); /*sti();*/ }while (0);
	//spin_unlock_irqrestore(&g_iommapSpinLock, flags);
	XSPIN_UNLOCK(g_iommapSpinLock, flags);
	return ENOMEM;


PassMem:
	DMSG(KERN_INFO "Virtual: 0x%08X 0x%08X, Physical: 0x%08X 0x%08X\n",
				(unsigned)Block[0].uHigh, (unsigned)Block[0].uLow, (unsigned)Block[1].uHigh, (unsigned)Block[1].uLow);


    return SUCCESSFUL;
}



/*
#ifdef _DEBUG
static BOOL TestGetPhysicalAddress(unsigned long memmapapiInfo)
{

	//virt_to_phys();
	void *virturlAddr;
	unsigned long physicalAddr=0;
	
	// try to kmalloc 1Kb
	virturlAddr = kmalloc(1024, GFP_KERNEL);

    if (!virturlAddr) {
        printk ("Kmalloc failed for virturlAddr\n");
        return ENOMEM;
    }

    // fill it with X
    memset (virturlAddr, 'X', 1024);

	physicalAddr = virt_to_phys((void*)virturlAddr);
	printk (KERN_ERR "physicalAddr 1: 0x%X\n", (unsigned)physicalAddr);

	physicalAddr = virt_to_phys((void*)virturlAddr+20);
	printk (KERN_ERR "physicalAddr 2: 0x%X\n", (unsigned)physicalAddr);

    if (virturlAddr)
        kfree (virturlAddr);


    return SUCCESSFUL;
}
#endif
*/




/********************************************************************************
 * Function	: driver_vma_open													*
 * Access	: Local																*
 * Abstract	: Increase driver count if driver mmap.								*
 ********************************************************************************/
#ifdef _DEBUG
static void driver_vma_open(struct vm_area_struct *vma)
{
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p)\n", __FUNCTION__, __LINE__, vma);

    IOMMAP_MODULE_INCREASE;
}

/********************************************************************************
 * Function	: driver_vma_close													*
 * Access	: Local																*
 * Abstract	: Decrease driver count if driver unmap.							*
 ********************************************************************************/
static void driver_vma_close(struct vm_area_struct *vma)
{
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p)\n", __FUNCTION__, __LINE__, vma);
	
    IOMMAP_MODULE_DECREASE;
}

/********************************************************************************
 * Function	: VMA_NOPAGE_WRAPPER												*
 * Access	: Local																*
 * Abstract	: Wrapper for no page access violation.								*
 ********************************************************************************/
static VMA_NOPAGE_WRAPPER(vma, address, write)
{
	struct vm_area_struct *vma_tmp = NULL;
    struct page *page = NOPAGE_SIGBUS;
	pgd_t *pgd = NULL;
	pmd_t *pmd = NULL;
	pte_t *ptep = NULL, pte;
	PUD_DEFINE

	DMSG("FUNCTION: %s, LINE: %d, ARG(%p)\n", __FUNCTION__, __LINE__, vma);

	// Unsupported OS version
	CHECK_UNKNOW_OS_ERROR_2();

	// abort if the MemMapTestGetPhysicalAddress function is not run.
	// this is already implement in library level to make this call.
	if(g_vma_offset == 0)
		return page;

	// abort if g_vma is invalid
	if(g_vma == NULL)
		return page;

	// abort if previous vma is not valid.
	vma_tmp = find_vma(g_vma->vm_mm, g_vma_offset);
	if(vma_tmp == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID VMA VALUE\n",__FUNCTION__, __LINE__);
		return page;
	}
	
	printk("<1>vma_tmp->vm_start 0x%.8lX vma_tmp->vm_end 0x%.8lX\n",vma_tmp->vm_start, vma_tmp->vm_end);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
	printk("<1>current->mm 0x%.8lX current->active_mm 0x%.8lX init_mm 0x%.8lX\n",current->mm, current->active_mm, init_mm);
	printk("<1>current->mm->pgd 0x%.8lX current->active_mm->pgd 0x%.8lX init_mm.pgd 0x%.8lX\n",current->mm->pgd, current->active_mm->pgd, init_mm.pgd);
#else
	printk("<1>current->mm 0x%.8lX current->active_mm 0x%.8lX\n",current->mm, current->active_mm);
	printk("<1>current->mm->pgd 0x%.8lX current->active_mm->pgd 0x%.8lX\n",current->mm->pgd, current->active_mm->pgd);
#endif

	// physical RAM address translation is done by walking thru the page table.
	// walking thru 3 level of page directory.
	pgd = pgd_offset(g_vma->vm_mm, g_vma_offset);

	if(pgd == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PGD VALUE\n",__FUNCTION__, __LINE__);
		return page;
	}

	PUD_OFFSET_WRAPPER(pgd, g_vma_offset);

	//pmd = pmd_offset(pgd, g_vma_offset);
	pmd = pmd_offset(PGD_OR_PUD, g_vma_offset);

	if(pmd == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PMD VALUE\n",__FUNCTION__, __LINE__);
		return page;
	}

	ptep = PTE_OFFSET_WRAPPER(pmd, g_vma_offset);

	if(ptep == NULL) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PPTE VALUE\n",__FUNCTION__, __LINE__);
		return page;
	}

	pte = *ptep;

	PTE_UNMAP_WRAPPER(ptep);


	if(!pte_present(pte)) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PTE VALUE\n",__FUNCTION__, __LINE__);
		return page;
	}

	page = pte_page(pte);

	if(page == NOPAGE_SIGBUS) {
		printk (KERN_ERR "FUNCTION:%s, LINE: %d, ERROR: INVALID PAGE VALUE\n",__FUNCTION__, __LINE__);
		return page;
	}

    get_page(page);

    return page;
}


/********************************************************************************
 * Structure declaration for vma												*
 ********************************************************************************/
static struct vm_operations_struct driver_vm_ops = {
    .open = driver_vma_open,
    .close = driver_vma_close,
    .nopage = driver_vma_nopage,
};


/********************************************************************************
 * Function	: driver_mmap														*
 * Access	: Global															*
 * Abstract	: driver map to user space.											*
 ********************************************************************************/
int driver_mmap(struct file *file, struct vm_area_struct *vma)
{
	DMSG("FUNCTION: %s, LINE: %d, ARG(%p,%p)\n", __FUNCTION__, __LINE__, file, vma);
	DMSG("<1>vma->vm_start 0x%.8lX vma->vm_end 0x%.8lX\n",vma->vm_start, vma->vm_end);

    vma->vm_ops = &driver_vm_ops;
	
    driver_vma_open(vma);

	driver_vma_nopage(vma, 0, 0);

    return SUCCESSFUL;
}
#endif
